package com.enoram.restify.server;

import java.io.IOException;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.google.gwt.user.server.rpc.RPCServletUtils;
import com.google.inject.Key;
import com.google.inject.name.Names;

public abstract class AbstractRestServiceServlet extends HttpServlet {

	private static final long serialVersionUID = 1L;

	protected transient ThreadLocal<HttpServletRequest> perThreadRequest;
	protected transient ThreadLocal<HttpServletResponse> perThreadResponse;

	
	@Override
	protected void doGet(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
		doPost(req, resp);
	}
	@Override
	protected void doPost(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		// Ensure the thread-local data fields have been initialized

		try {
			// Store the request & response objects in thread-local storage.
			//
			synchronized (this) {
				validateThreadLocalData();
				perThreadRequest.set(request);
				perThreadResponse.set(response);
			}

			processPost(request, response);

		} catch (Throwable e) {
			// Give a subclass a chance to either handle the exception or
			// rethrow it
			//
			doUnexpectedFailure(e);
		} finally {
			// null the thread-locals to avoid holding request/response
			//
			perThreadRequest.set(null);
			perThreadResponse.set(null);
		}

	}

	protected abstract void processPost(HttpServletRequest request,
			HttpServletResponse response) throws Throwable;

	/**
	 * Override this method to control what should happen when an exception
	 * escapes the {@link #doPost} method. The default implementation will log
	 * the failure and send a generic failure response to the client.
	 * <p>
	 * An "expected failure" is an exception thrown by a service method that is
	 * declared in the signature of the service method. These exceptions are
	 * serialized back to the client, and are not passed to this method. This
	 * method is called only for exceptions or errors that are not part of the
	 * service method's signature, or that result from SecurityExceptions,
	 * SerializationExceptions, or other failures within the RPC framework.
	 * <p>
	 * Note that if the desired behavior is to both send the GENERIC_FAILURE_MSG
	 * response AND to rethrow the exception, then this method should first send
	 * the GENERIC_FAILURE_MSG response itself (using getThreadLocalResponse),
	 * and then rethrow the exception. Rethrowing the exception will cause it to
	 * escape into the servlet container.
	 * 
	 * @param e
	 *            the exception which was thrown
	 */
	protected void doUnexpectedFailure(Throwable e) {
		try {
			getThreadLocalResponse().reset();
		} catch (IllegalStateException ex) {
			/*
			 * If we can't reset the request, the only way to signal that
			 * something has gone wrong is to throw an exception from here. It
			 * should be the case that we call the user's implementation code
			 * before emitting data into the response, so the only time that
			 * gets tripped is if the object serialization code blows up.
			 */
			throw new RuntimeException("Unable to report failure", e);
		}
		ServletContext servletContext = getServletContext();
		RPCServletUtils.writeResponseForUnexpectedFailure(servletContext,
				getThreadLocalResponse(), e);
	}

	/**
	 * Gets the <code>HttpServletResponse</code> object for the current call. It
	 * is stored thread-locally so that simultaneous invocations can have
	 * different response objects.
	 */
	protected final HttpServletResponse getThreadLocalResponse() {
		synchronized (this) {
			validateThreadLocalData();
			return perThreadResponse.get();
		}
	}
	  /**
	   * Gets the <code>HttpServletRequest</code> object for the current call. It is
	   * stored thread-locally so that simultaneous invocations can have different
	   * request objects.
	   */
	  protected final HttpServletRequest getThreadLocalRequest() {
	    synchronized (this) {
	      validateThreadLocalData();
	      return perThreadRequest.get();
	    }
	  }
	/**
	 * Initializes the perThreadRequest and perThreadResponse fields if they are
	 * null. This will occur the first time they are accessed after an instance
	 * of this class is constructed or deserialized. This method should be
	 * called from within a 'synchronized(this) {}' block in order to ensure
	 * that only one thread creates the objects.
	 */
	private void validateThreadLocalData() {
		if (perThreadRequest == null) {
			perThreadRequest = new ThreadLocal<HttpServletRequest>();
		}
		if (perThreadResponse == null) {
			perThreadResponse = new ThreadLocal<HttpServletResponse>();
		}
	}


}
