package com.enoram.restify.server;

import com.google.gson.Gson;
import com.google.gwt.user.client.rpc.IncompatibleRemoteServiceException;
import com.google.gwt.user.client.rpc.RemoteService;
import com.google.gwt.user.client.rpc.RpcToken;
import com.google.gwt.user.client.rpc.SerializationException;
import com.google.gwt.user.client.rpc.impl.AbstractSerializationStream;
import com.google.gwt.user.server.rpc.RPCRequest;
import com.google.gwt.user.server.rpc.RPCServletUtils;
import com.google.gwt.user.server.rpc.SerializationPolicy;
import com.google.gwt.user.server.rpc.UnexpectedException;
import com.google.gwt.user.server.rpc.impl.LegacySerializationPolicy;
import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamReader;
import com.google.gwt.user.server.rpc.impl.ServerSerializationStreamWriter;
import com.google.gwt.user.server.rpc.impl.TypeNameObfuscator;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URLEncoder;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

public final class REST {

	private static Logger logger = Logger.getLogger(REST.class.getName());
	/**
	 * Maps primitive wrapper classes to their corresponding primitive class.
	 */
	private static final Map<Class<?>, Class<?>> PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS = new HashMap<Class<?>, Class<?>>();

	/**
	 * Static map of classes to sets of interfaces (e.g. classes). Optimizes
	 * lookup of interfaces for security.
	 */
	private static Map<Class<?>, Set<String>> serviceToImplementedInterfacesMap;

	private static final HashMap<String, Class<?>> TYPE_NAMES;

	static {
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Boolean.class,
				Boolean.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Byte.class, Byte.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Character.class,
				Character.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Double.class,
				Double.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Float.class, Float.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Integer.class,
				Integer.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Long.class, Long.TYPE);
		PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.put(Short.class, Short.TYPE);

		TYPE_NAMES = new HashMap<String, Class<?>>();
		TYPE_NAMES.put("Z", boolean.class);
		TYPE_NAMES.put("B", byte.class);
		TYPE_NAMES.put("C", char.class);
		TYPE_NAMES.put("D", double.class);
		TYPE_NAMES.put("F", float.class);
		TYPE_NAMES.put("I", int.class);
		TYPE_NAMES.put("J", long.class);
		TYPE_NAMES.put("S", short.class);

		serviceToImplementedInterfacesMap = new HashMap<Class<?>, Set<String>>();
	}

	public static RestRequest decodeRequest(RestRequestContent requestContent,
			String serviceIntfName) {

		if (requestContent == null) {
			throw new NullPointerException("encodedRequest cannot be null");
		}

		if (requestContent.isEmptyApiNamed()) {
			throw new IllegalArgumentException(
					"Rest Request Info cannot have empty Class Name. Check for /_api/.. path");
		}

		ClassLoader classLoader = Thread.currentThread()
				.getContextClassLoader();

		Class<?> serviceIntf;

		try {

			serviceIntf = getClassFromSerializedName(serviceIntfName,
					classLoader);

		} catch (ClassNotFoundException e) {
			throw new IncompatibleRemoteServiceException(
					"Could not locate requested interface '" + serviceIntfName
							+ "' in default classloader", e);
		}

		String apiNamed = requestContent.getApiNamed();

		// METHOD
		// Parse Method using apiNamed to do the lookup
		String serviceMethodPath = requestContent.getMethodPath();

		Method method = RESTAPI.getMethod(apiNamed, serviceMethodPath);

		// METHOD's PARAMS
		// here we need to parse for Param Values from REST call
		String[] params = RESTAPI.getMethodParams(apiNamed, serviceMethodPath);

		RestParam restParam = requestContent.getRestParam();

		// The output of method params will be in String[] {"@id int",
		// "@start int", "@len int"}
		Object[] parameterValues = new Object[params.length];
		int index = 0;
		for (String param : params) {
			// for each param, we will find that param's value from the
			// queryString
			Object value = getParamValue(param, requestContent);
			if(value==null){
				throw new IncompatibleRemoteServiceException(
						"Could not locate required param value for '" + param);
			}
			parameterValues[index++] = value;

		}

		return new RestRequest(requestContent.getApiNamed(), method,
				parameterValues);

	}

	/**
	 * Get the value from
	 * <code>requestContent.restParam<code/> query string value or from requestContent.content if 
	 * this is object
	 * 
	 * @param param
	 *            "@id int" for named param;
	 *            "class com.enoram.restify.client.Comment" for Class param
	 * @param restParam
	 * @return
	 */
	private static Object getParamValue(String param,
			RestRequestContent requestContent) {

		final RestParam restParam = requestContent.getRestParam();
	
		ParamParser parser = new ParamParser(param);
		Object objectValue=null;
		
		ObjectReturnedParamCallBack callback = new ObjectReturnedParamCallBack(objectValue, requestContent) ;
					
		parser.parse(callback);
		
		return callback.getResult();

	}	
	
	private static Object getParamValueXX(String param,
			RestRequestContent requestContent) {

		//remove the "class " word
		param = param.replaceAll("class ", "");
		
		String[] ss = param.split(" ");

		if (ss.length < 2) {
			throw new IllegalAccessError(
					"Bad param. Expected 'name type' pair param");
		}

		if (ss[0].startsWith("@")) {
			RestParam restParam = requestContent.getRestParam();
			String named = ss[0];
			named = named.substring(1); // remove @
			try {

				// need to get the Object value
				String type = ss[1]; // int, long, etc
				Object objectValue = restParam.deserializeValue(type, named);

				return objectValue;

			} catch (Exception e) {
				throw new RuntimeException("Parameter " + param
						+ " of is of an unknown value '" + ss[1] + "'", e);
			}
		} else {

			try {
				Class clazz = Class.forName(ss[1]);
				Gson gson = new Gson();
				Object jsonObject = gson.fromJson(requestContent.getContent(),
						clazz);

				return jsonObject;

			} catch (ClassNotFoundException e) {
				throw new RuntimeException("Parameter " + param
						+ " of is of an unknown value '" + ss[1] + "'", e);
			}

		}

	}

	private static boolean isNativeEntityType(Class fieldType) {
		if (fieldType == Integer.TYPE || fieldType == Integer.class
				|| fieldType == String.class || fieldType == Date.class) {
			return true;
		}
		return false;
	}

	/**
	 * Returns a default serialization policy.
	 * 
	 * @return the default serialization policy.
	 */
	public static SerializationPolicy getDefaultSerializationPolicy() {
		return LegacySerializationPolicy.getInstance();
	}

	/**
	 * Returns the {@link Class} instance for the named class or primitive type.
	 * 
	 * @param serializedName
	 *            the serialized name of a class or primitive type
	 * @param classLoader
	 *            the classLoader used to load {@link Class}es
	 * @return Class instance for the given type name
	 * @throws ClassNotFoundException
	 *             if the named type was not found
	 */
	public static Class<?> getClassFromSerializedName(String serializedName,
			ClassLoader classLoader) throws ClassNotFoundException {
		Class<?> value = TYPE_NAMES.get(serializedName);
		if (value != null) {
			return value;
		}

		return Class.forName(serializedName, false, classLoader);
	}

	public static String invokeAndEncodeResponse(Object target,
			Method serviceMethod, Object[] args,
			SerializationPolicy serializationPolicy)
			throws SerializationException {
		return invokeAndEncodeResponse(target, serviceMethod, args,
				serializationPolicy, AbstractSerializationStream.DEFAULT_FLAGS);
	}

	public static String invokeAndEncodeResponse(Object target,
			Method serviceMethod, Object[] args,
			SerializationPolicy serializationPolicy, int flags)
			throws SerializationException {
		if (serviceMethod == null) {
			throw new NullPointerException("serviceMethod");
		}

		if (serializationPolicy == null) {
			throw new NullPointerException("serializationPolicy");
		}

		String responsePayload;
		try {
			Object result = serviceMethod.invoke(target, args);

			responsePayload = encodeResponseForSuccess(serviceMethod, result,
					serializationPolicy, flags);
		} catch (IllegalAccessException e) {
			SecurityException securityException = new SecurityException(
					formatIllegalAccessErrorMessage(target, serviceMethod));
			securityException.initCause(e);
			throw securityException;
		} catch (IllegalArgumentException e) {
			SecurityException securityException = new SecurityException(
					formatIllegalArgumentErrorMessage(target, serviceMethod,
							args));
			securityException.initCause(e);
			throw securityException;
		} catch (InvocationTargetException e) {
			// Try to encode the caught exception
			//
			Throwable cause = e.getCause();

			responsePayload = encodeResponseForFailure(serviceMethod, cause,
					serializationPolicy, flags);
		}

		return responsePayload;
	}

	public static String encodeResponseForSuccess(Method serviceMethod,
			Object object, SerializationPolicy serializationPolicy, int flags)
			throws SerializationException {
		if (serviceMethod == null) {
			throw new NullPointerException("serviceMethod cannot be null");
		}

		if (serializationPolicy == null) {
			throw new NullPointerException("serializationPolicy");
		}

		Class<?> methodReturnType = serviceMethod.getReturnType();
		if (methodReturnType != void.class && object != null) {
			Class<?> actualReturnType;
			if (methodReturnType.isPrimitive()) {
				actualReturnType = getPrimitiveClassFromWrapper(object
						.getClass());
			} else {
				actualReturnType = object.getClass();
			}

			if (actualReturnType == null
					|| !methodReturnType.isAssignableFrom(actualReturnType)) {
				throw new IllegalArgumentException(
						"Type '"
								+ printTypeName(object.getClass())
								+ "' does not match the return type in the method's signature: '"
								+ getSourceRepresentation(serviceMethod) + "'");
			}
		}

		return encodeResponse(methodReturnType, object, false, flags,
				serializationPolicy);
	}

	/**
	 * Returns the {@link java.lang.Class Class} for a primitive type given its
	 * corresponding wrapper {@link java.lang.Class Class}.
	 * 
	 * @param wrapperClass
	 *            primitive wrapper class
	 * @return primitive class
	 */
	private static Class<?> getPrimitiveClassFromWrapper(Class<?> wrapperClass) {
		return PRIMITIVE_WRAPPER_CLASS_TO_PRIMITIVE_CLASS.get(wrapperClass);
	}

	/**
	 * Straight copy from
	 * {@link com.google.gwt.dev.util.TypeInfo#getSourceRepresentation(Class)}
	 * to avoid runtime dependency on gwt-dev.
	 */
	private static String printTypeName(Class<?> type) {
		// Primitives
		//
		if (type.equals(Integer.TYPE)) {
			return "int";
		} else if (type.equals(Long.TYPE)) {
			return "long";
		} else if (type.equals(Short.TYPE)) {
			return "short";
		} else if (type.equals(Byte.TYPE)) {
			return "byte";
		} else if (type.equals(Character.TYPE)) {
			return "char";
		} else if (type.equals(Boolean.TYPE)) {
			return "boolean";
		} else if (type.equals(Float.TYPE)) {
			return "float";
		} else if (type.equals(Double.TYPE)) {
			return "double";
		}

		// Arrays
		//
		if (type.isArray()) {
			Class<?> componentType = type.getComponentType();
			return printTypeName(componentType) + "[]";
		}

		// Everything else
		//
		return type.getName().replace('$', '.');
	}

	/**
	 * Returns the source representation for a method signature.
	 * 
	 * @param method
	 *            method to get the source signature for
	 * @return source representation for a method signature
	 */
	private static String getSourceRepresentation(Method method) {
		return method.toString().replace('$', '.');
	}

	public static String encodeResponseForFailure(Method serviceMethod,
			Throwable cause, SerializationPolicy serializationPolicy, int flags)
			throws SerializationException {
		if (cause == null) {
			throw new NullPointerException("cause cannot be null");
		}

		if (serializationPolicy == null) {
			throw new NullPointerException("serializationPolicy");
		}

		if (serviceMethod != null
				&& !RPCServletUtils.isExpectedException(serviceMethod, cause)) {
			throw new UnexpectedException("Service method '"
					+ getSourceRepresentation(serviceMethod)
					+ "' threw an unexpected exception: " + cause.toString(),
					cause);
		}

		return encodeResponse(cause.getClass(), cause, true, flags,
				serializationPolicy);
	}

	/**
	 * Returns a string that encodes the results of an RPC call. Private
	 * overload that takes a flag signaling the preamble of the response
	 * payload.
	 * 
	 * For REST we will use GSON to return the String notation of this result object.
	 * 
	 * The JSON response String will be prefixed by "//OK//" or "//EX//"
	 * 
	 * 
	 * @param object
	 *            the object that we wish to send back to the client
	 * @param wasThrown
	 *            if true, the object being returned was an exception thrown by
	 *            the service method; if false, it was the result of the service
	 *            method's invocation
	 * @return a string that encodes the response from a service method
	 * @throws SerializationException
	 *             if the object cannot be serialized
	 */
	private static String encodeResponse(Class<?> responseClass, Object object,
			boolean wasThrown, int flags,
			SerializationPolicy serializationPolicy)
			throws SerializationException {

		
		String response = null;
		
		if(wasThrown){
			return encodeResponseRPC(responseClass, object, wasThrown, flags, serializationPolicy);
		}
		
		if (responseClass != void.class) {
			Gson gson = new Gson();
			
			response = gson.toJson(object);
			
		}

		String bufferStr = (wasThrown ? "//EX//" : "") + response;
		return bufferStr;
	}
	private static String encodeResponseRPC(Class<?> responseClass, Object object,
			boolean wasThrown, int flags,
			SerializationPolicy serializationPolicy)
			throws SerializationException {

		ServerSerializationStreamWriter stream = new ServerSerializationStreamWriter(
				serializationPolicy);
		stream.setFlags(flags);

		stream.prepareToWrite();
		if (responseClass != void.class) {
			stream.serializeValue(object, responseClass);
		}

		String bufferStr = (wasThrown ? "//EX" : "//OK") + stream.toString();
		return bufferStr;
	}
	private static String formatIllegalAccessErrorMessage(Object target,
			Method serviceMethod) {
		StringBuffer sb = new StringBuffer();
		sb.append("Blocked attempt to access inaccessible method '");
		sb.append(getSourceRepresentation(serviceMethod));
		sb.append("'");

		if (target != null) {
			sb.append(" on target '");
			sb.append(printTypeName(target.getClass()));
			sb.append("'");
		}

		sb.append("; this is either misconfiguration or a hack attempt");

		return sb.toString();
	}

	private static String formatIllegalArgumentErrorMessage(Object target,
			Method serviceMethod, Object[] args) {
		StringBuffer sb = new StringBuffer();
		sb.append("Blocked attempt to invoke method '");
		sb.append(getSourceRepresentation(serviceMethod));
		sb.append("'");

		if (target != null) {
			sb.append(" on target '");
			sb.append(printTypeName(target.getClass()));
			sb.append("'");
		}

		sb.append(" with invalid arguments");

		if (args != null && args.length > 0) {
			sb.append(Arrays.asList(args));
		}

		return sb.toString();
	}

	private static String formatMethodNotFoundErrorMessage(
			Class<?> serviceIntf, String serviceMethodName,
			Class<?>[] parameterTypes) {
		StringBuffer sb = new StringBuffer();

		sb.append("Could not locate requested method '");
		sb.append(serviceMethodName);
		sb.append("(");
		for (int i = 0; i < parameterTypes.length; ++i) {
			if (i > 0) {
				sb.append(", ");
			}
			sb.append(printTypeName(parameterTypes[i]));
		}
		sb.append(")'");

		sb.append(" in interface '");
		sb.append(printTypeName(serviceIntf));
		sb.append("'");

		return sb.toString();
	}
	public static String formatServletPath(String path){
		if(!path.endsWith("/")){
			return path + "/";
		}
		
		return path;
	}
	
	public static String encoded(String param){
		try {
			return URLEncoder.encode(param,"UTF-8");
		} catch (UnsupportedEncodingException e) {
			logger.severe("Error encoded " +param);
			e.printStackTrace();
		}
		
		return null;
	}
}
